//-----------------------------------------------------------------
// ImageFilter.cs (c) 2007 by Charles Petzold for "Beautiful Code"
//-----------------------------------------------------------------
using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Reflection;
using System.Reflection.Emit;
using System.Runtime.InteropServices;

class ImageFilter
{
    double[] filter;
    int cxFilter;
    int cyFilter;

    public ImageFilter(int cxFilter, double[] filter)
    {
        this.filter = filter;
        this.cxFilter = cxFilter;
        this.cyFilter = filter.Length / cxFilter;
    }

    public bool ApplyFilter(Bitmap bmp, bool willGenerateCode)
    {
        // Sprawdzenie czy mapa bitowa jest we waciwym formacie.
        if (bmp.PixelFormat != PixelFormat.Format24bppRgb &&
            bmp.PixelFormat != PixelFormat.Format32bppArgb &&
            bmp.PixelFormat != PixelFormat.Format32bppPArgb &&
            bmp.PixelFormat != PixelFormat.Format32bppRgb)
        {
            return false;
        }

        Rectangle rect = new Rectangle(0, 0, bmp.Width, bmp.Height);
        BitmapData data = bmp.LockBits(rect, ImageLockMode.ReadOnly, 
                                        bmp.PixelFormat);

        // Skopiowanie pikseli z mapy bitowej do tablicy src.
        int cBytes = data.Stride * data.Height;
        byte[] src = new byte[cBytes];
        Marshal.Copy(data.Scan0, src, 0, cBytes);
        bmp.UnlockBits(data);

        // Alokacja tablicy docelowej.
        byte[] dst = new byte[cBytes];

        // Uruchomienie metody filtrujcej.
        if (!willGenerateCode)
            FilterMethodOne(src, dst, Math.Abs(data.Stride), Math.Abs(data.Stride) / data.Width);
        else
            FilterMethodTwo(src, dst, Math.Abs(data.Stride), Math.Abs(data.Stride) / data.Width);

        // Skopiowanie pikseli z tablicy dst do mapy bitowej.
        data = bmp.LockBits(rect, ImageLockMode.WriteOnly, bmp.PixelFormat);
        Marshal.Copy(dst, 0, data.Scan0, cBytes);
        bmp.UnlockBits(data);

        return true;
    }

    // Proste podejcie C#.
    void FilterMethodOne(byte[] src, byte[] dst, int stride, int bytesPerPixel)
    {
        int cBytes = src.Length;
        int cFilter = filter.Length;

        for (int iDst = 0; iDst < cBytes; iDst++)
        {
            double pixelsAccum = 0;
            double filterAccum = 0;

            for (int iFilter = 0; iFilter < cFilter; iFilter++)
            {
                int yFilter = iFilter / cyFilter;
                int xFilter = iFilter % cxFilter;

                int iSrc = iDst + stride * (yFilter - cyFilter / 2) + 
                                    bytesPerPixel * (xFilter - cxFilter / 2);

                if (iSrc >= 0 && iSrc < cBytes)
                {
                    pixelsAccum += filter[iFilter] * src[iSrc];
                    filterAccum += filter[iFilter];
                }
            }

            if (filterAccum != 0)
                pixelsAccum /= filterAccum;

            dst[iDst] = pixelsAccum < 0 ? (byte)0 : (pixelsAccum > 255 ? 
                                            (byte)255 : (byte)pixelsAccum);
        }
    }

    // Generowanie jzyka poredniego w locie.
    void FilterMethodTwo(byte[] src, byte[] dst, int stride, int bytesPerPixel)
    {
        int cBytes = src.Length;

        DynamicMethod dynameth = new DynamicMethod("Go", typeof(void),
            new Type[] { typeof(byte[]), typeof(byte[]) }, GetType());

        ILGenerator generator = dynameth.GetILGenerator();

        // Definicje zmiennych lokalnych.
        generator.DeclareLocal(typeof(int));        // Index 0 = iDst
        generator.DeclareLocal(typeof(double));     // Index 1 = pixelsAccum
        generator.DeclareLocal(typeof(double));     // Index 2 = filterAccum

        // ustawienie iDst na zero.
        generator.Emit(OpCodes.Ldc_I4_0);
        generator.Emit(OpCodes.Stloc_0);

        // Definicja etykiety na pocztku ptli iDst.
        Label labelTop = generator.DefineLabel();
        generator.MarkLabel(labelTop);

        // Ustawienie pixelAccum i filterAccum na zero.
        generator.Emit(OpCodes.Ldc_R8, 0.0);
        generator.Emit(OpCodes.Dup);
        generator.Emit(OpCodes.Stloc_1);
        generator.Emit(OpCodes.Stloc_2);

        // Iteracja przez tablic filtra.
        for (int iFilter = 0; iFilter < filter.Length; iFilter++)
        {
            // Pominicie tego, jeli element filtru ma warto 0.
            if (filter[iFilter] == 0)
                continue;

            int xFilter = iFilter % cxFilter;
            int yFilter = iFilter / cxFilter;
            int offset = stride * (yFilter - cyFilter / 2) +
                        bytesPerPixel * (xFilter - cxFilter / 2);

            // Umieszczenie tablicy src na stosie.
            generator.Emit(OpCodes.Ldarg_0);

            // Definicja etykiet do rozgaziania.
            Label labelLessThanZero = generator.DefineLabel();
            Label labelGreaterThan = generator.DefineLabel();
            Label labelLoopBottom = generator.DefineLabel();

            generator.Emit(OpCodes.Ldloc_0);        // indeks dst na stosie
            generator.Emit(OpCodes.Ldc_I4, offset); // przesunicie na stosie
            generator.Emit(OpCodes.Add);            // Dodawanie
            generator.Emit(OpCodes.Dup);            // Dwukrotne tworzenie duplikatu.
            generator.Emit(OpCodes.Dup);

            // Sprawdzenie czy mniej ni zero (usuwa pierwszy indeks)
            generator.Emit(OpCodes.Ldc_I4_0);
            generator.Emit(OpCodes.Blt_S, labelLessThanZero);

            // Sprawdzenie czy wicej ni rozmiar mapy bitowej (usuwa drugi indeks)
            generator.Emit(OpCodes.Ldc_I4, cBytes);
            generator.Emit(OpCodes.Bge_S, labelGreaterThan);
            
            // Dostp do tablicy src i konwersja na typ double
            generator.Emit(OpCodes.Ldelem_U1);
            generator.Emit(OpCodes.Conv_R8);

            // Specjalne przetwarzanie, jeli element filtru ma warto 1.
            if (filter[iFilter] == 1)
            {
                // Element src jest na stosie, a wic nie robimy nic.
            }
            else if (filter[iFilter] == -1)
            {
                // Negacja elementu src na stosie.
                generator.Emit(OpCodes.Neg);
            }
            else
            {
                // Mnoenie przez element filtru i kumulacja.
                generator.Emit(OpCodes.Ldc_R8, filter[iFilter]);
                generator.Emit(OpCodes.Mul);
            }

            // Kumulacja pixelsAccum z liczb na stosie.
            generator.Emit(OpCodes.Ldloc_1);
            generator.Emit(OpCodes.Add);
            generator.Emit(OpCodes.Stloc_1);

            // Akumulacja filterAccum.
            generator.Emit(OpCodes.Ldc_R8, filter[iFilter]);
            generator.Emit(OpCodes.Ldloc_2);    // filterAccum
            generator.Emit(OpCodes.Add);
            generator.Emit(OpCodes.Stloc_2);
            generator.Emit(OpCodes.Br, labelLoopBottom);

            // Czyszczenie na dole ptli filtru.
            generator.MarkLabel(labelLessThanZero);
            generator.Emit(OpCodes.Pop);
            generator.MarkLabel(labelGreaterThan);
            generator.Emit(OpCodes.Pop);
            generator.Emit(OpCodes.Pop);
            generator.MarkLabel(labelLoopBottom);
        }

        generator.Emit(OpCodes.Ldarg_1);        // tablica dst
        generator.Emit(OpCodes.Ldloc_0);        // indeks iDst

        // Definicja kilku etykiet dla celw rozgaziania.
        Label labelSkipDivide = generator.DefineLabel();
        Label labelCopyQuotient = generator.DefineLabel();
        Label labelBlack = generator.DefineLabel();
        Label labelWhite = generator.DefineLabel();
        Label labelDone = generator.DefineLabel();

        // Podzielenie pixelsAccum przez filterAccum.
        generator.Emit(OpCodes.Ldloc_1);        // pixelsAccum
        generator.Emit(OpCodes.Ldloc_2);        // filterAccum
        generator.Emit(OpCodes.Dup);            // Zrobienie kopii.
        generator.Emit(OpCodes.Ldc_R8, 0.0);    // Wstawienie 0 na stos.

        // Uniknicie dzielenia przez zero.
        generator.Emit(OpCodes.Beq_S, labelSkipDivide);
        generator.Emit(OpCodes.Div);            // Dzielenie.
        generator.Emit(OpCodes.Br_S, labelCopyQuotient);

        generator.MarkLabel(labelSkipDivide);
        generator.Emit(OpCodes.Pop);            // Pop filterAccum

        generator.MarkLabel(labelCopyQuotient);
        generator.Emit(OpCodes.Dup);            // Zrobienie kopii ilorazu.
        generator.Emit(OpCodes.Dup);            // Dodanie jeszcze jednej.

        // Sprawdzenie czy iloraz jest mniejszy od zera.
        generator.Emit(OpCodes.Ldc_R8, 0.0);
        generator.Emit(OpCodes.Blt_S, labelBlack);

        // Sprawdzenie czy iloraz jest wikszy od 255.
        generator.Emit(OpCodes.Ldc_R8, 255.0);
        generator.Emit(OpCodes.Bgt_S, labelWhite);

        // Konwersja wartoci na stosie na bajt.
        generator.Emit(OpCodes.Conv_U1);
        generator.Emit(OpCodes.Br_S, labelDone);

        // Jeli < 0, ustawienie jej na bajt zerowy.
        generator.MarkLabel(labelBlack);
        generator.Emit(OpCodes.Pop);
        generator.Emit(OpCodes.Pop);
        generator.Emit(OpCodes.Ldc_I4_S, 0);
        generator.Emit(OpCodes.Br_S, labelDone);

        // Jeli > 255, ustawienie na 255.
        generator.MarkLabel(labelWhite);
        generator.Emit(OpCodes.Pop);
        generator.Emit(OpCodes.Ldc_I4_S, 255);
        
        // W tym momencie warto docelowa mapy bitowej jest na stosie.
        generator.MarkLabel(labelDone);

        // Zapisujemy j w tablicy.
        generator.Emit(OpCodes.Stelem_I1);

        // Zwikszenie iDst.
        generator.Emit(OpCodes.Ldloc_0);    // Wstawienie iDst na stos.
        generator.Emit(OpCodes.Ldc_I4_1);   // Wstawienie 1 na stos.
        generator.Emit(OpCodes.Add);        // Dodanie 1 do iDst.
        generator.Emit(OpCodes.Dup);        // Duplikacja.
        generator.Emit(OpCodes.Stloc_0);    // Zapisanie wyniku w iDst.

        generator.Emit(OpCodes.Ldc_I4, cBytes);  // Wstawienie wartoci cBytes na stos.
        generator.Emit(OpCodes.Blt, labelTop);   // Przejcie do gry, jeli iDst < cBytes.

        // Wszystko zrobione  powrt.
        generator.Emit(OpCodes.Ret);

        // Wywoanie metody.
        dynameth.Invoke(this, new object[] { src, dst });
    }
}

